#include<algorithm>
#include<iomanip>
#include<iostream>
#include<string>
#include<fstream>
#include<sstream>
#include<vector>
#include<cmath>
using namespace std;

class Read{
public:
Read() : mism(0), chr("0"), strand('+'), pos(0) {};
Read(string ch, char st, long int p, int m) : mism(m), chr(ch), strand(st), pos(p) {};
string chr;
char strand;
long int pos;
int mism;
};

void parse_options(int argc, char* argv[], string &inp, string &inp2, string &ref, 
	string &read_cov, string &fdr_thresh, string &meth_call)
{
	int res = 0;

	long int i, j;
	for(i = 1; i < argc; i++){
		if(argv[i][0] != '-'){
			cerr << "ERROR: Invalid char=" << argv[i][1] << "=this" << endl; 
			exit(0); 
			return ;
		}
		switch(argv[i][1]){
		case 'g': ref = argv[++i]; break;//file with a single reference in fasta format
		case 'r': read_cov = argv[++i]; break;//threshold on read coverage
		case 'f': fdr_thresh = argv[++i]; break;//threshold on fdr 0.1 means 10% of reads having this error are false positive
		case 'm': inp = argv[++i]; break;//file with correct meth level: ref, strand, pos(starts 1), count of Cs, count of (Cs+Ts)
		case 't': inp2 = argv[++i]; break;//file with meth level in BRAT-BW format (output of BRAT-bw or BRAT-nova)
		case 'c': meth_call = argv[++i]; break;//percentage of sequencing errors from total bases
		default: cerr << "ERROR: Invalid char=" << argv[i][1] << "=this" << endl; exit(0); break;
		}
	}//for i

	return ;
}//parse_opt()

void merge_errors(vector<double> &errors, vector<double> &current_errors){
	long int asize = errors.size();
	vector<double> temp_errors(asize, 0.0);
	long int cur_size = current_errors.size();
	long int i = 0, j = 0, u = 0;
	while((i < asize) && (j < cur_size) && (u < asize)){
		if(errors[i] <= current_errors[j])
		{
			temp_errors[u] = errors[i];
			u++;
			i++;
		}
		else{
			temp_errors[u] = current_errors[j];
			j++;
			u++;
		}
	}//while
	for(i = 0; i < asize; i++)
		errors[i] = temp_errors[i];
}//merge_errors

int main(int argc, char* argv[]){

	string inp ;//file with correct meth call: chr, strand, pos(starts with 1), Cs, (Cs+Ts)
	string inp2 ;//coverage of ACGT in bismark format: read_id, +/- for meth/unmeth, chr, pos(starts with 1), char
	string ref ;//reference to know positions
	//string outp = argv[4];// distribution of errors at each pos (correct coverage >= 10, and calculated coverage >= 5)
	string read_cov ;
	string fdr_thresh ;//p-value percentage threshold 0.1 means 10% of all errors
	string meth_call;

	parse_options(argc, argv, inp, inp2, ref, read_cov, fdr_thresh, meth_call);
	srand( time(0) );

	int thresh_cov = 10;
	if(!read_cov.empty()){
		istringstream istr_cov(read_cov);
		istr_cov >> thresh_cov;
	}
	double thresh_fdr = 0.2;
	if(!fdr_thresh.empty()){
		istringstream istr_call(fdr_thresh);
		istr_call >> thresh_fdr;//fdr threshold
	}
	double thresh_meth_call2 = 0.5;
	if(!meth_call.empty()){
		istringstream istr_meth(meth_call);
		istr_meth >> thresh_meth_call2;
	}

//	ofstream out;
//	out.open(outp.c_str(), ios::out);
	if(inp.empty())
	{
		cout << "ERROR: please provide file with correct methylation level" << endl;
		exit(0);
	}
	if(inp2.empty()){
		cout << "ERROR: please provide file with BRAT-nova's acgt-count's output" << endl;
		exit(0);	
	}
	if(ref.empty()){
		cout << "ERROR: please provide file with the reference in fasta format that was used to generate reads to estimate methylation call accuracy" << endl;
		exit(0);	
	}
	ifstream in;
 	in.open(ref.c_str(), ios::in);
	if(!in){
		cout << "ERROR: can't open file " << ref << endl;
		exit(0);
	}

	long int  TP = 0, FP = 0, TN = 0, FN = 0;
//	ofstream out;
//	out.open(outp.c_str(), ios::out);


	string header, line;
	getline(in, header);//starting with >
	istringstream istr4(header);
	string correct_chr;
	istr4 >> correct_chr;
	correct_chr = correct_chr.substr(1, correct_chr.length() - 1);

	long int length = 0;	
	getline(in, line);
	
	while(!in.eof()){
		length += line.length();
		getline(in, line);
	}//while
	in.close(); in.clear();

   

	vector< pair<double, int> > tool_meth(length);
	for(long int i = 0; i < length; i++){
		tool_meth[i].first = 0.0;
		tool_meth[i].second = 0;
	}//for i

	in.open(inp2.c_str(), ios::in);
	string chr, cov, read_id, dum;
	double meth_perc = 0.0;
	char strand;
	long int start, stop;

	//correct meth level
	long int total_chr = 0;
	in >> read_id;
	long int covCs = 0, cov_total = 0, pos;
	while(!in.eof()){
		in >> strand >> chr >> pos >> dum;
		pos--;
		if(chr == correct_chr){
			total_chr++;
			tool_meth[pos].second += 1;
			if(strand == '+')
				tool_meth[pos].first += 1 ;//number of Cs
		}
		in >> read_id;
	}//while
	in.close(); in.clear();

	for(long int i = 0; i < length; i++){
		tool_meth[i].first = ( 0.0 + tool_meth[i].first) / (0.0 + tool_meth[i].second) ;
	}//for i
cout << "methylation level from aligner is calculated; total entries with correct chrom = " 
	<< total_chr <<endl;

	vector< pair<double, int> > meth_level(length);
	for(long int i = 0; i < length; i++){
		meth_level[i].first = 0.0;
		meth_level[i].second = 0;
	}//for i

	in.open(inp.c_str(), ios::in);



	//correct meth level
	in >> chr;

	while(!in.eof()){
		in >> strand >> pos >> covCs >> cov_total;
		pos--;
		meth_perc = (covCs + 0.0) / (0.0 + cov_total);
		meth_level[pos].first = meth_perc;
		meth_level[pos].second = cov_total;

		in >> chr;
	}//while
cout << "Meth level original is read" <<endl;

	in.close(); in.clear();

 	long int tool_covered = 0;
	for(int i = 0; i < length; i++)
	{
		if(tool_meth[i].second >= thresh_cov)
			tool_covered++;
	}
	
	long int top_pvalue = thresh_fdr * 100.0;
	long int rest = 100 - top_pvalue;
	vector<double> errors(top_pvalue*tool_covered , 0.0);
	long int ind_errors = 0;
	for(int j = 0; j < top_pvalue ; j++){
		
		for(long int i = 0;  i < length; i++){
			long int cur_cov = tool_meth[i].second;
			if(cur_cov >= thresh_cov){
			   long int cur_Cs = rand() % (cur_cov + 1);
			   double cur_meth = (cur_Cs + 0.0)/(0.0 + cur_cov);
			   double cur_error = abs( meth_level[i].first - cur_meth);
			   errors[ind_errors] = cur_error;
			   ind_errors++;				
			}//if coverage is good
		}//for i
//cout << "round j = " << j << endl;
	}//for j
	
	sort(errors.begin(), errors.end());

	vector<double> current_errors(tool_covered, 0.0);
	
	for(int j = 0; j < rest ; j++){
		ind_errors = 0;
		for(long int i = 0;  i < length; i++){
			long int cur_cov = tool_meth[i].second;
			if(cur_cov >= thresh_cov){
			   long int cur_Cs = rand() % (cur_cov + 1);
			   double cur_meth = (cur_Cs + 0.0)/(0.0 + cur_cov);
			   double cur_error = abs( meth_level[i].first - cur_meth);
			   current_errors[ind_errors] = cur_error;
			   ind_errors++;				
			}//if coverage is good
		}//for i
		sort(current_errors.begin(), current_errors.end());
		
		merge_errors(errors, current_errors);
//cout << "round j = " << j << endl;
	}//for j
//	cout << "errors size = " << ind_errors << ", 10% FDR index = " << top_pvalue << endl;
	double fdr = errors[errors.size() - 1] ;
	cout << "Methylation level error corresponding to the given FDR threshold is  = " << fdr << endl;


	errors.clear();
	errors.resize(0);

	long int rand_correct = 0;
	long int total_covered = 0;
	long int total_correct = 0;
	for(long int i = 0; i < length; i++){
		start = i;
		int int_cov  = tool_meth[i].second;
		//if correct coverage >= 10 and calculated coverage is >= 5
		if((meth_level[start].second >= thresh_cov) && (int_cov >= thresh_cov)){
			total_covered++;
			meth_perc = tool_meth[start].first;
			double abs_error = (meth_level[start].first - meth_perc);
			if(abs(abs_error) <= fdr){
				rand_correct++;
			}
//			out << std::fixed << std::setprecision(12) << abs_error << endl;
			if(meth_level[start].first <  thresh_meth_call2){
				if(meth_perc < thresh_meth_call2)
				{
					total_correct++;
					TN++;
				}//real unmeth and fake unmeth
				else
					FP++;
			}//unmeth
			else{
				if(meth_perc >= thresh_meth_call2){
					total_correct++;	
					TP++;

				}	
				else
					FN++;	
			}//else meth
			
		}//if coverage
	}//for i

	cout << "Methylation call threshold = " << thresh_meth_call2 << endl;
	cout << "FDR threshold = " << thresh_fdr << endl;
	cout << "Total covered bases >= " << thresh_cov <<" is : " << total_covered << endl;
	cout << "Total correctly called methylation status: " << total_correct << endl;
	double map_acc = 100*(total_correct + 0.0)/(0.0 + total_covered);
	cout << "Methylation call accuracy: " << map_acc << endl;	
	double rand_acc = 100*(rand_correct + 0.0)/(0.0 + total_covered);
	cout << "Methylation level accuracy for the given FDR threshold = " << rand_acc << endl;
	cout << "TP, FP, TN, FN (calculated for the methylation call threshold): " << endl;
	cout << TP << "\t" << FP << "\t" << TN << "\t" << FN << endl;
	double recall = (0.0 + TP)/(0.0 + TP + FP);
	double FDR = (0.0 + FP ) / (0.0 + TP + FP);
	cout << "recall , FDR (calculated for the methylation call threshold): " << recall << "\t" << FDR << endl;




}
