#include<algorithm>
#include<ctime>
#include<fstream>
#include<string>
#include<vector>
#include<iostream>
#include<sstream>
using namespace std;

class Read{
public:
	Read(string r, string ch, char str, long int p) : indel_pos(0), indel_len(0), mism(0), read(r), chr(ch), strand(str), pos(p) {};
	Read() : read(""), chr("") , strand('+'), pos(0), indel_len(0), indel_pos(0), mism(0) {};
	
	string read;
	string chr;//chrom name
	char strand ;
	long int pos;
	short int indel_type;//0 is insertion, 1 is deletion
	int indel_pos;//after which indel is introduced within clipped read
	int indel_len;//length of indel
	int mism;//number of mismatches introduced in mappable bases (excluding insertion bases)
	string orig_read;//includes indel, but no mismatches
};



char change_base(char ch){
vector<char> noA(3);
noA[0] = 'C';
noA[1] = 'G';
noA[2] = 'T';
vector<char> noT(3);
noT[0] = 'A';
noT[1] = 'C';
noT[2] = 'G';
vector<char> noC(3);
noC[0] = 'A';
noC[1] = 'G';
noC[2] = 'T';
vector<char> noG(3);
noG[0] = 'A';//can't be A since in rev-compl A in a read can map to G in genome
noG[1] = 'C';
noG[2] = 'T';
		int base = rand() % 3;
                char next = ch;
                if(next == 'A')
                       return noA[base];
                else if(next == 'C')
                       return noC[base];
                else if(next == 'G')
                       return noG[base];
                else
                       return noT[base];
}//mism


void add_bs_mism(vector<Read> &reads, double perc_bs){

	long int asize = reads.size();
	long int total_Cs = 0;
	for(long int i = 0; i < asize; i++){
		int read_len = reads[i].read.length();
		for(int j = 0; j < read_len; j++){
			if(reads[i].read[j] == 'C')
				total_Cs++;
		}//for j
	}//for i

	long int total_change = (perc_bs / 100.0) * total_Cs;

	long int cur_change = 0;
	while(cur_change < total_change){
		long int ind = rand() % asize;
		int read_len = reads[ind].read.length();
		for(int j = 0; j < read_len ; j++){
			if(reads[ind].read[j] == 'C'){
				reads[ind].read[j] = 'T';
				reads[ind].orig_read[j] = 'T';
				cur_change++;
			}//if
		}//for
		
	}//while
	cout << "finished bisulfite-conversion" << endl;
}//bs_mism

void rev_compl(string &s){
        int len = s.length();
        stringstream ss;
        for(int i = len - 1; i >= 0; i--){
                if(s[i] == 'A' || s[i] == 'a')
                        ss << 'T';
                else if(s[i] == 'C' || s[i] == 'c' )
                        ss << 'G';
                else if(s[i] == 'G' || s[i] == 'g')
                        ss << 'C';
                else if(s[i] == 'T' || s[i] == 't')
                        ss << 'A';
                else
                        ss << 'N';
        }//for
        s = ss.str();
}//rev_compl

void add_indel(vector<Read> &art_reads, const long int &read_index, const int &pos_within_read, const int & indel_length, const int & type, const int &max_indel){

	
	if(type == 0){//insertion
		string insert("");
		stringstream ss;
		for(int i = 0; i < indel_length; i++){
			int base = rand() % 4;
			if(base == 0)
				ss << 'A';
			else if(base == 1)
				ss << 'C';
			else if(base == 2)
				ss << 'G';
			else
				ss << 'T';
		}//generate random insert
		insert = ss.str();
		
		int orig_read_len = art_reads[read_index].read.length() - max_indel;

		art_reads[read_index].read = (art_reads[read_index]).read.substr(0, pos_within_read + 1) + insert + (art_reads[read_index]).read.substr(pos_within_read + 1, art_reads[read_index].read.length() - pos_within_read - 1);	

		//clip extra bases
		int cur_len = orig_read_len + max_indel + indel_length;
		int clip_extra = art_reads[read_index].read.length() - orig_read_len;
		art_reads[read_index].read = art_reads[read_index].read.substr(clip_extra, cur_len - clip_extra);

		//adjust position
		if(art_reads[read_index].strand == '+'){
			art_reads[read_index].pos += clip_extra;
		}//if positive
		else{//negative
			//no adjustment needed: we trimmed front of read, whose rev-complement is mapped
			//to pos, and the resulting read's rev-complement the leftmost end of alignment
			//does not change

		}//if negative

		//set indel information
		art_reads[read_index].indel_type = 0;//insert
		art_reads[read_index].indel_len = indel_length;
		art_reads[read_index].indel_pos = pos_within_read - clip_extra;
	}//insertion	
	else{
		int cur_length = art_reads[read_index].read.length();
		int orig_length = cur_length - max_indel;
		int start = pos_within_read + indel_length + 1;
		art_reads[read_index].read = art_reads[read_index].read.substr(0, pos_within_read + 1) + art_reads[read_index].read.substr(start, cur_length - start);

		//clip extra bases
		cur_length = art_reads[read_index].read.length();
		int clip_extra = cur_length - orig_length;
		if(clip_extra > 0){
			art_reads[read_index].read = art_reads[read_index].read.substr(clip_extra, cur_length - clip_extra); 
		}
		//adjust position
		if(art_reads[read_index].strand == '+')
			art_reads[read_index].pos += clip_extra;
                art_reads[read_index].indel_type = 1;//deletion
                art_reads[read_index].indel_len = indel_length;
                art_reads[read_index].indel_pos = pos_within_read - clip_extra;


	}//deletion
}//add_indel

void clip_indel(vector<Read> &reads, int indel){

	long int asize = reads.size();
	//indels were not introduced, so read length is same
	int read_len = reads[0].read.length();
	for(long int i = 0; i < asize; i++){
		reads[i].read = reads[i].read.substr(indel, read_len - indel);
		if(reads[i].strand == '+')
			reads[i].pos += indel; 
	}//for
}
void set_orig_reads(vector<Read> &reads){
	//after add_indels and after clip_indel
	long int asize = reads.size();
	for(long int i = 0; i < asize; i++)
		reads[i].orig_read = reads[i].read;
}//set_orig_reads

void add_sequence_errors(vector<Read> &reads, const long int &total_mism, const double &start, const double &stop){
	//start * read_len  is inclusive position
	long int asize = reads.size();
	for(long int i = 0; i < total_mism; i++){

		//choose randomly a read
		long int index = rand() % asize;
		int start2 = start * reads[index].read.length();
		int stop2 = stop * reads[index].read.length();
		//choose randomly position
		int pos = ( rand() % (stop2 - start2)) + start2;
		reads[index].read[pos] = change_base(reads[index].read[pos]);	

	}//for i
}//add_sequence_errors

void add_adapter(vector<Read> &reads, const long int &total_reads, const string & adapter, const int &max_adapter_len){
	int max_len = max_adapter_len;
	long int asize = reads.size();
        vector<long int> rand_index(asize);
        for(long int i = 0; i < asize; i++){
        	rand_index[i] = i;         
	
	}//for i
	//without replacement
	long int stop = asize;
	int read_len = reads[0].read.length();
	for(long int i = 0; i < total_reads; i++){

        	int read_len = reads[i].read.length();
		//randomly choose a read
		long int index = rand() % stop;
		long int read_index = rand_index[index];
		rand_index[index] = rand_index[stop - 1];
		stop--;

		int ad_len = rand() % max_len + 1;
		
		reads[read_index].read = reads[read_index].read.substr(0, read_len - ad_len) + adapter.substr(0, ad_len);
		
	}//
	

}//add_adapter

void count_mism(vector<Read> &reads){
	long int asize = reads.size();
	for(long int i = 0; i < asize; i++){
		int mism = 0;

//check if indel
if(reads[i].indel_len > 0){
//cout << "count_mism: indel_len = " << reads[i].indel_len << " > 0" << endl;
		if(reads[i].indel_type == 1){
			//if deletion, then compare char-by-char orig_read and read
			int read_len = reads[i].read.length();
			for(int j = 0; j < read_len; j++){
				if(reads[i].read[j] != reads[i].orig_read[j])
					mism++;		
			}//for j			
		}//if deletion
		else{
			int read_len = reads[i].read.length();
			int stop = reads[i].indel_pos + 1;
			for(int j = 0; j < stop; j++){
                                if(reads[i].read[j] != reads[i].orig_read[j])
                                        mism++;
				
			}//for j
			stop += reads[i].indel_len;
                        for(int j = stop; j < read_len; j++){
                                if(reads[i].read[j] != reads[i].orig_read[j])
                                        mism++;
                               
                        }//for j

		}//if insertion
}//if indel
else{//not indel
                        int read_len = reads[i].read.length();
                        for(int j = 0; j < read_len; j++){
                                if(reads[i].read[j] != reads[i].orig_read[j])
                                        mism++;
                        }//for j

	
}


		reads[i].mism = mism;
	}//for i

}//count_mism

void parse_options(int argc, char* argv[], string &inp, string &outp, string &outp_pos, 
	string &radius, string &totals, string &perc_se, string &perc_snp, 
	string &perc_ind, string &perc_adapter, string &max_adapter, string &outp_methC)
{
	int res = 0;

	long int i, j;
	for(i = 1; i < argc; i++){
		if(argv[i][0] != '-'){
			cerr << "ERROR: Invalid char=" << argv[i][1] << "=this" << endl; 
			exit(0); 
			return ;
		}
		switch(argv[i][1]){
		case 'g': inp = argv[++i]; break;//file with paths to files with references
		case 'r': outp = argv[++i]; break;//reads in fastq
		case 'p': outp_pos = argv[++i]; break;//positions
		case 'l': radius = argv[++i]; break;//read length
		case 't': totals = argv[++i]; break;//total reads
		case 'e': perc_se = "1"; break;//percentage of sequencing errors from total bases
		case 's': perc_snp = atoi(argv[++i]); break;//percentage of SNPs from total bases
		case 'i': perc_ind = atoi(argv[++i]); break;//percentage of reads with an indel
		case 'a': perc_adapter = true; break;//percentage of reads having an adapter sequence at 3'end
		case 'm': max_adapter = 1; break;//max length of adapter sequence
		case 'o': outp_methC = argv[++i]; break;//output methylation level of Cs 
		default: cerr << "ERROR: Invalid char=" << argv[i][1] << "=this" << endl; exit(0); break;
		}
	}//for i

	return ;
}//parse_opt()

int main(int argc, char* argv[]){
	srand( time(0) );

	string inp;//argv1
	string outp;//argv2
	string outp_pos;//argv3
	string radius;//argv4
	string totals; //argv5
	string outp_methC;//output meth level


	double perc_bs = 97;
	double perc_seq_err = 2, perc_SNPs = 1, perc_indels = 1, perc_adapter_cont = 10, max_adapter_len = 15;
	string perc_se, perc_snp, perc_ind, perc_adapter, max_adapter;

	parse_options(argc, argv, inp, outp, outp_pos, radius, totals, perc_se, perc_snp, perc_ind, perc_adapter, max_adapter, outp_methC);

	if(!perc_se.empty()){
		istringstream istr_se(perc_se);
		istr_se >> perc_seq_err;
		if(perc_seq_err < 0 || perc_seq_err > 10)
			perc_seq_err = 2;
	}
	if(!perc_snp.empty()){
		istringstream istr_snp(perc_snp);
		istr_snp >> perc_SNPs;
		if(perc_SNPs < 0 || perc_SNPs > 10)
			perc_SNPs = 1;
	}
	if(!perc_ind.empty()){
		istringstream istr_ind(perc_ind);
        istr_ind >> perc_indels;
		if(perc_indels < 0 || perc_indels > 100)
			perc_indels = 1;
	}
	if(!perc_adapter.empty()){
		istringstream istr1(perc_adapter);
		istr1 >> perc_adapter_cont;//this one is percentage of reads having adapter sequence
		if(perc_adapter_cont < 0 || perc_adapter_cont > 100)
			perc_adapter_cont = 10;
	}
	if(!max_adapter.empty()){
		istringstream istr2(max_adapter);
		istr2 >> max_adapter_len;
		if(max_adapter_len < 0 || max_adapter_len > 50)
			max_adapter_len = 15;
	}

	int orig_read_len = 100;
	if(!radius.empty()){
		istringstream istr(radius);
		istr >> orig_read_len;//length of reads
		if(orig_read_len < 0 || orig_read_len > 1000)
			orig_read_len = 100;
	}
	long int indel = 0;//max length of indels
	int read_len = orig_read_len ;
	if(!perc_ind.empty()){//if indels
		indel = 10;//max length of indel
		read_len += indel;//so that deletion does not cause shortening
	}


	
	long int total = 100;
	if(!totals.empty()){
		istringstream istr3(totals);
		istr3 >> total;
		if(total < 0 || total > 50000000)
			total = 100;
	}
	
	if(inp.empty()){
		cout << "ERROR: please provide a fasta file with a reference (option -g)." << endl;
		exit(0);
	}
	if(outp.empty())
		outp = "reads.fastq";
	if(outp_pos.empty())
		outp_pos = "positions.txt";
	if(outp_methC.empty())
		outp_methC = "methylation_level.txt";
	//read file and find out the length of chrom
	ifstream in;
	string line, header, qual;

	long int read_id = 0;
	long int area = 0;
	long int total_bs_mism = 0;	

	vector<Read> art_reads;
	art_reads.reserve(total);

 	in.open(inp.c_str(), ios::in);

	getline(in, header);//starting with >
	istringstream istr4(header);
	string first;
	istr4 >> first;
	first = first.substr(1, first.length() - 1);

	getline(in, line);
	stringstream orig;
	while(!in.eof()){
		orig << line;
		getline(in, line);
	}//while
	in.close(); in.clear();

    string ss = orig.str();//single reference sequence
	long int length = ss.length();	
	vector<short int> meth_perc;	
	meth_perc.resize(length, 0);

	//set randomly methylation percentage at each C and G
	for(long int i = 0; i < length; i++){
		if(ss[i] == 'C' || ss[i] == 'G'){
			int mperc = ((rand() % 4) + 1) * 20;//20%, 405, 60%, 80% of methylated Cs at this pos
			meth_perc[i] = mperc;
		}//if
	}//for

	vector< pair< short int, short int > > meth_level;
	meth_level.resize(length);
	for(long int i = 0; i < length; i++){
		meth_level[i].first = 0;
		meth_level[i].second = 0;
	}//for i

	long int indel = 0;//max length of indels
	int read_len = orig_read_len ;
	if(!perc_ind.empty()){//if indels
		indel = 3;//fixed length of indel
		read_len += indel;//so that deletion does not cause shortening of read length
	}// if indels

	long int totalC = 0, totalG = 0;
	long int total_Cm = 0;//methylated
	long int total_Gm = 0;//methylated
	for(long int j = 0; j < total; j++){
			
		long int pos = rand() % (length - read_len);
		string sub = ss.substr(pos, read_len);
		int astrand = rand() % 2;
		char strand = '+';
		if(astrand == 1){
			rev_compl(sub);
			strand = '-';
		}
		//convert Cs to Ts
		if(strand == '+'){
			long int stop = pos + read_len;
			long int read_ind = indel;
			for(long int e = pos + indel; e < stop; e++){//do not change C->T in clipped indel part
				if(ss[e] == 'C'){
					totalC++;
					meth_level[e].second++;
					//decide to methylate or not
					int is_meth = (rand() % 100) + 1;
					if(is_meth > meth_perc[e]){
						//C is not methylated at this position, so convert it to T
						sub[read_ind] = 'T';
					}//if
					else{
						total_Cm++;
						meth_level[e].first++;
					}//else methylated
				}//if C
				read_ind++;
			}//for e to set meth for C
		}//+
		else{
			long int stop = pos + read_len - 1 - indel;//inclusive
			long int read_ind = indel;
			for(long int e = stop; e >= pos; e--){
				if(ss[e] == 'G'){
					totalG++;
					meth_level[e].second++;
					int is_meth = (rand() % 100) + 1;
					if(is_meth > meth_perc[e]){
						sub[read_ind] = 'T';
					}//if
					else{
						total_Gm++;
						meth_level[e].first++;
					}//else
				}//if G
				read_ind++;
			}//for e to set meth for G
		
		}//-

		//end convert Cs to Ts
		Read aread(sub, first, strand, pos + 1);//pos count starts with 1
		art_reads.push_back(aread);

	}// for j: generate reads from this reference


cout << "reads are extracted from references" << endl;
	//calculate bs-rate
	double bsrateC = 100.0*((0.0 + totalC - total_Cm)/(0.0 + totalC));
	double bsrateG = 100.0*((0.0 + totalG - total_Gm)/(0.0 + totalG));
//	cout << "BS-rate for + strand: ratio of unmethylated Cs to total Cs in reads = " << bsrateC <<endl;
//	cout << "BS-rate for - strand: ratio of unmethylated Cs to total Cs in reads = " << bsrateG <<endl;

	//output true methylation level 
	ofstream out_methC;
	out_methC.open(outp_methC.c_str(), ios::out);
	for(long int i = 0; i < length; i++){
		if(ss[i] == 'C'){
			out_methC << first << "\t" << '+' << "\t" << (i + 1) 
				<< "\t" << meth_level[i].first << "\t" << meth_level[i].second << endl;
		}//if
		else if(ss[i] == 'G'){
			out_methC << first << "\t" << '-' << "\t" << (i + 1) 
				<< "\t" << meth_level[i].first << "\t" << meth_level[i].second << endl;
		}//if
	}//for i
	out_methC.close(); out_methC.clear();


	//	Step1.	Introduce indels 
	if(!perc_ind.empty()){
		//count the number of reads with indels
		long int total_reads_indels = (perc_indels / 100.0) * art_reads.size() ;//of reads
		if(total_reads_indels > art_reads.size())
			total_reads_indels = art_reads.size();
		cout << "total number of reads with indels: " << total_reads_indels << endl;
		long int asize = art_reads.size();
		vector<long int> rand_indels(asize);
		for(long int i = 0; i < asize; i++){
			rand_indels[i] = i;
		}//for i
		long int stop = total_reads_indels ;
		int no_indel = 5;//no indels in these first and last bases of the read
		for(long int i = 0; i < total_reads_indels; i++){
			//randomly select a read
			long int rread_ind = rand() % stop;
			//randomly select type: insertion or deletion
			int type = rand() % 2;
			int indel_length = indel;//fixed length
			int pos_within_read = ( rand() % (orig_read_len - indel - indel - no_indel - no_indel )) + indel + indel + no_indel;
			add_indel(art_reads, rand_indels[rread_ind], pos_within_read, indel_length, type, indel);
			//swap indices so that not to reuse the same
			long int last = rand_indels[stop - 1];
			rand_indels[rread_ind] = last;
			stop--;			
		}//for i				
	}//if indels
//no need "clip_indels" since if no indels read_length is orig_read_length
//	else//clip extra bases from front of reads, adjust positions
//		clip_indels(art_reads, indel);
	
	set_orig_reads(art_reads);

	//	Step2. Introduce SNPs: uniform random distribution
	long int total_snp = (perc_SNPs / 100.0) * art_reads.size() * orig_read_len;
cout << "Total SNPs = " << total_snp << endl;
	add_sequence_errors(art_reads, total_snp, 0.0, 1.0);

	//	Step3. Convert Cs to Ts using perc_bs
	//Already added bs mismatches
//	add_bs_mism(art_reads, perc_bs);

	//	Step4. Add adapter sequences
	if(!perc_adapter.empty()){
		string adapter = "CGGTTCAGCAGGAATGCCGAGATCGGAAGAGCGGTTCAGCAGGAATGCCG";
		long int total_adapters = (perc_adapter_cont / 100.0) * art_reads.size() ; 
		if(max_adapter_len > adapter.length())
			max_adapter_len = adapter.length();
		add_adapter(art_reads, total_adapters, adapter, max_adapter_len);
	}//if adapters


	//	Step 5. Introduce sequencing errors: in the first 2/3 of the read x mismatches
	//		in the next 1/6 of the read 10x mismatches, and in the next 1/6 100x mismatches

	//calculate total number of mismatches
	long int total_mism = art_reads.size() * orig_read_len * (perc_seq_err / 100.0);
cout << "total sequencing errors = " << total_mism << endl;
	long int first_mism = total_mism / (28.5);
	long int second_mism = first_mism * 2.5;
	long int third_mism = first_mism * 25;

	double stop1 =  (2.0 / 3.0);
	double stop2 = (5.0 / 6.0);

	add_sequence_errors(art_reads, first_mism, 0.0, stop1);
	add_sequence_errors(art_reads, second_mism, stop1, stop2);
        add_sequence_errors(art_reads, third_mism, stop2, 1.0); 

	count_mism(art_reads);


        ofstream out;//reads
        out.open(outp.c_str(), ios::out);
        ofstream out_pos;
        out_pos.open(outp_pos.c_str(), ios::out);
	long int asize = art_reads.size();
	read_id = 1;
	string head("@");

if(perc_ind.empty()){//qual of the same length as read_len
        stringstream ss_qual;
        for(int i = 0; i < orig_read_len; i++){
                ss_qual << 'J';
        }
        qual = ss_qual.str();
}

	for(long int i = 0; i < asize; i++){
		stringstream ss_header;
		ss_header << head;
		ss_header << read_id;

if(!perc_ind.empty()){		
	int aread_len = art_reads[i].read.length();
        stringstream ss_qual;
        for(int i = 0; i < aread_len; i++){
                ss_qual << 'J';
        }
        qual = ss_qual.str();
}

		string anid = ss_header.str();
		out << anid << endl << art_reads[i].read << endl << "+" << endl << qual << endl;
		out_pos << read_id << "\t" << art_reads[i].read << "\t" 
			<< art_reads[i].chr << "\t" << art_reads[i].strand 
			<< "\t" << art_reads[i].pos << "\t" << art_reads[i].mism 
			<< "\t" << art_reads[i].indel_len << "\t" << art_reads[i].indel_pos << endl;
		read_id++;
	}
	out_pos.close();
	out.close();
	return 0;

}
