#include<iostream>
#include<string>
#include<fstream>
#include<sstream>
#include<vector>
using namespace std;


typedef unsigned long int _uli;
typedef unsigned int _ui;
typedef unsigned short int _usi;

class Read{
public:
Read() : mism(0), indel_len(0), chr("0"), strand('+'), pos(0) {};
Read(string ch, char st, long int p, int m) : indel_len(0), mism(m), chr(ch), strand(st), pos(p) {};
string chr;
char strand;
long int pos;
int mism;
int indel_len;
};

void make_mask(vector<_uli> &mask){
	_uli SIZE = 64;
        _uli one = 1;
        for(int i = 0; i <= SIZE; i++){
                _uli res = (one << i);
                mask.push_back(res);
        }//for
}//make mask

void count_mism(const string & cigar, const string &md, int & mism, int &indel_len){
	mism = 0; 
	indel_len = 0;
	//break cigar into chunks ending with character
	vector<int> indices;
	int asize = cigar.length();
	for(int i = 0; i < asize; i++){
		if(!isdigit(cigar[i]))
			indices.push_back(i);	
	}//for	
	int ind_size = indices.size();
	//check if there is D, deletion 
	bool isD = false;
	vector<int> deletions;

	for(int i = 0; i < ind_size; i++){
		if(cigar[indices[i]] == 'D' || cigar[indices[i]] == 'I'){
			int cur_index = indices[i];
			//find deletion length
			int prev_ind = 0;
			if(i > 0){
				prev_ind = indices[i - 1] + 1;//index of indel len
			}
				string sub = cigar.substr(prev_ind, cur_index - prev_ind);
				istringstream istr(sub);
				int alen ;
				istr >> alen;
				indel_len += alen;
				if(cigar[indices[i]] == 'D')
					deletions.push_back(alen);
		}//if
	}//for

	int j = 0, del_ind = 0;
	asize = md.length();
	while(j < asize){
		
		if(md[j] == '^'){
			int alen = deletions[del_ind];
			del_ind++;
			j += alen;	
		}//if deletion
		else{
			if(!isdigit(md[j]))
				mism++;
		}
		j++;
	}//scan md
}//count_mism

void adjust_pos(long int &pos, const string &cigar){
	//find clip_left
	//find the first H character
	int i = 0;
	int asize = cigar.length();
	while(i < asize){
		if(!isdigit(cigar[i]))
			break;
		i++;
	}//while

	if(i < asize){
		//check what is the first character
		if(cigar[i] == 'H' || cigar[i] == 'S'){
			string sub = cigar.substr(0, i);
			istringstream istr(sub);
			int clip_left ;
			istr >> clip_left;
			pos -= clip_left;//in BS-seeker
		}//if
	}
}//adjust_pos

void parse_options(int argc, char* argv[], string &inp_pos, string &outp, string &totals, 
	string &pos_thresh)
{
	int res = 0;

	long int i, j;
	for(i = 1; i < argc; i++){
		if(argv[i][0] != '-'){
			cerr << "ERROR: Invalid char=" << argv[i][1] << "=this" << endl; 
			exit(0); 
			return ;
		}
		switch(argv[i][1]){
		case 'r': outp = argv[++i]; break;//output of brat-bw
		case 'p': inp_pos = argv[++i]; break;//original positions
		case 't': totals = argv[++i]; break;//total reads attempted to map
		case 'e': pos_thresh = argv[++i]; break;//threshold for position to be considered correct
		default: cerr << "ERROR: Invalid char=" << argv[i][1] << "=this" << endl; exit(0); break;
		}
	}//for i

	return ;
}//parse_opt()

int main(int argc, char* argv[]){
	string inp ;//file with correct positions: read_id, read, chr, strand, pos, mism, indel_len, indel_pos
	string inp2 ;//output of brat-bw
	string totals ;//total reads attempted to ma
	string thresh_pos ;

	parse_options(argc, argv, inp, inp2, totals, thresh_pos);
	long int thresh = 50;
	if(!thresh_pos.empty()){
		istringstream istr_pos(thresh_pos);
		istr_pos >> thresh ;
	}
	

	long int total = 100;
	if(totals.empty()){
		cout << "ERROR: please provide the total number of reads attempted to map (option t)" << endl;
		exit(0);
	}
	else{
		istringstream istr(totals);
		istr >> total;
	}
	ifstream in;
	in.open(inp.c_str(), ios::in);
	
	if(!in){
		cout << "ERROR: please provide file with original positions (option p)" << endl;
		exit(0);
	}


	string read, chr;
	char strand;
	int indel_len, indel_pos, mism;
	long int id, st = 0, en = 0, pos, last ;
	vector<Read> reads;//max len 76
	reads.resize(total+2);
	in >> id;	
	while(!in.eof()){
		in >> read >> chr >> strand >> pos >> mism >> indel_len >> indel_pos;
		
		reads[id].chr = chr;
		reads[id].pos = pos;
		reads[id].strand = strand;
		reads[id].mism = mism;
		reads[id].indel_len = indel_len;
		in >> id;
	
	}//while
	in.close(); in.clear();
	cout << "\n Positions are loaded size = " << reads.size() << endl;


	vector<_uli> mask;
	make_mask(mask);

	long int total_pos_good = 0;
	in.open(inp2.c_str(), ios::in);

	if(!in){
		cout << "ERROR: please provide output file in SAM format (option p)" << endl;
		exit(0);
	}

	long int total_wrong = 0, total_good = 0, wrong_better = 0, wrong_equal = 0, wrong_worse = 0;
	_uli flag = 0, zero = 0;
	int temp;
	string idd, qual, cigar, md, dum;
        
	long int total_chr = 0, total_strand = 0, total_pos = 0, total_pos_delta = 0;
	long int counter = 0;
	in >> idd;//id starts with 0 and pos starts with 0 
    while(!in.eof()){

		if(!(idd[0] == '@')){
			istringstream istr2(idd);
			istr2 >> id; 
			counter++;
			if(counter > total + 2)
			{
				cout << "ERROR counter > total" << endl;
				break;
			}                
			in >> flag >> chr >>  pos >> temp >> cigar >> dum >> temp >> temp >> read >> qual ;

			getline(in, dum);
			adjust_pos(pos, cigar);		

			strand = '+';
			if((mask[4] & flag) > zero)
				strand = '-'; 
			if(reads[id].chr == chr && reads[id].strand == strand && reads[id].pos == pos)
			{		
				total_good++;
				if(strand == '+')
					total_pos_good++;	
			}//if	
			else
			{	total_wrong++;
				if(reads[id].chr != chr)
					total_chr++;
				else if(reads[id].strand != strand)
					total_strand++;
				else{			
				
					long int minus_pos = pos - thresh;
					long int pos_plus = pos + thresh;
					if(reads[id].pos >= minus_pos && reads[id].pos <= pos_plus)
						total_pos_delta++;
					else 
						total_pos++;
				}
			}//else
		}//if not @
		else{
			string line;
			getline(in, line);
		}
		in >> idd;

    }//while
	cout << "Total of correctly mapped reads = " << total_good << ", of them positive: " 
		<< total_pos_good <<  endl;

    long int total_reads = total_good + total_wrong;
    long int total_correct2 = total_reads - (total_chr + total_strand + total_pos);
	long int total_wrong2 = total_chr + total_strand + total_pos;
    double map_acc1 = (0.0 + total_good)/(0.0 + total_good + total_wrong);
    double map_acc2 = (0.0 + total_correct2)/(0.0 + total_reads);

    cout << "Mapping accuracy = " << (100.0 * map_acc2) << endl;
 
	in.close();
}
